# setup environment
import os
smoke_test = ('CI' in os.environ)  # for continuous integration tests

import scanpy as sc
import torch
import torch.nn as nn
import math
import numpy as np
import anndata as ad
import torch
import random
import pandas as pd
from sklearn.preprocessing import LabelEncoder
from torch.utils.data import Dataset,DataLoader,Sampler
from scipy.sparse import csr_matrix

def data_load(anndata, n_top_genes, batch_key = 'batch_2', condition_key = 'batch', groundtruth_key = None):
    ######### 该代码用于预处理scanpy的原始数据，将单细胞测序数据进行清洗和质控，并且normalized它们 ##########
    datasets = anndata.copy()

    datasets.obs['batch'] = datasets.obs[condition_key].astype('category')
    datasets.obs['batch_2'] = datasets.obs[batch_key].astype('category')
    if groundtruth_key is not None:
        datasets.obs['groundtruth'] = datasets.obs[groundtruth_key].astype('category')
    datasets.layers['counts'] = datasets.X.copy()
    # 数据筛选
    sc.pp.filter_cells(datasets, min_genes=200)
    sc.pp.filter_genes(datasets, min_cells=3)
    raw = datasets.copy()
    # 数据标准化
    sc.pp.normalize_total(datasets, target_sum=1e4)
    # 进行log1p转换
    sc.pp.log1p(datasets)
    datasets.layers['lognorm'] = datasets.X.copy()
    # 识别高变基因
    sc.pp.highly_variable_genes(datasets, batch_key= None, n_top_genes=n_top_genes)
    # 将高变基因存在var中
    datasets.raw = datasets
    # 保留高变基因
    datasets = datasets[:, datasets.var.highly_variable]
    # scale数据
    sc.pp.scale(datasets, max_value=10)

    # batch = np.unique(datasets.obs[batch_key])
    # list_batch = []
    # for x in batch:
    #     adata_iter = datasets[datasets.obs[batch_key] == x]
    #     sc.pp.normalize_total(adata_iter, target_sum=1e5)
    #     list_batch.append(adata_iter)

    # datasets = ad.concat(list_batch)
    # datasets = sc.pp.log1p(datasets, copy = True)
    # datasets.layers['lognorm'] = datasets.X.copy()
    # # 识别高变基因
    # sc.pp.highly_variable_genes(datasets, n_top_genes=n_top_genes, batch_key='batch')
    # # 将高变基因存在var中
    # datasets.raw = datasets
    # # 保留高变基因
    # datasets = datasets[:, datasets.var.highly_variable]

    var_names = datasets.var_names
    index_names = datasets.obs_names
    
    return datasets,raw,var_names,index_names

# class BatchDataLoader(object):
#     """
#     This custom DataLoader serves mini-batches that are either fully-observed (i.e. labeled)
#     or partially-observed (i.e. unlabeled) but never mixed.
#     """

#     def __init__(self, data_x, data_y, batch_size, num_classes, missing_label=-1):
#         super().__init__()
#         self.data_x = data_x
#         self.data_y = data_y
#         self.batch_size = batch_size
#         self.num_classes = num_classes

#         self.unlabeled = torch.where(data_y == missing_label)[0]
#         self.num_unlabeled = self.unlabeled.size(0)
#         self.num_unlabeled_batches = math.ceil(self.num_unlabeled / self.batch_size)

#         self.labeled = torch.where(data_y != missing_label)[0]
#         self.num_labeled = self.labeled.size(0)
#         self.num_labeled_batches = math.ceil(self.num_labeled / self.batch_size)

#         self.classed = [torch.where(data_y == c)[0] for c in range(num_classes)]
#         self.num_classed = [c.size(0) for c in self.classed]
#         self.num_class_batches = [math.ceil(indices / self.batch_size) for indices in self.num_classed]

#         assert self.data_x.size(0) == self.data_y.size(0)
#         assert len(self) > 0

#     @property
#     def size(self):
#         return self.data_x.size(0)

#     def __len__(self):
#         return self.num_unlabeled_batches + self.num_labeled_batches

#     def _sample_batch_indices(self):
#         batch_order = torch.randperm(len(self)).tolist()
#         unlabeled_idx = self.unlabeled[torch.randperm(self.num_unlabeled)]
#         labeled_idx = self.labeled[torch.randperm(self.num_labeled)]
#         class_idx = [self.classed[c][torch.randperm(self.num_classed[c])] for c in range(self.num_classes)]

#         # print(class_idx[0].shape,class_idx[1].shape)

#         slices = []

#         for i in range(self.num_unlabeled_batches):
#             _slice = unlabeled_idx[i * self.batch_size : (i + 1) * self.batch_size]
#             slices.append((_slice, False))

#         # for i in range(self.num_labeled_batches):
#         #     _slice = labeled_idx[i * self.batch_size : (i + 1) * self.batch_size]
#         #     slices.append((_slice, True))

#         for c in range(self.num_classes):
#             for i in range(self.num_class_batches[c]):
#                 _slice = class_idx[c][i * self.batch_size : (i + 1) * self.batch_size]
#                 slices.append((_slice, True))

#         return slices, batch_order

#     def __iter__(self):
#         slices, batch_order = self._sample_batch_indices()

#         for i in range(len(batch_order)):
#             _slice = slices[batch_order[i]]
#             if _slice[1]:
#                 # labeled
#                 yield self.data_x[_slice[0]], nn.functional.one_hot(
#                     self.data_y[_slice[0]], num_classes=self.num_classes
#                 )
#             else:
#                 # unlabeled
#                 yield self.data_x[_slice[0]], None

# class BatchDataLoader2(object):
#     """
#     This custom DataLoader serves mini-batches that are either fully-observed (i.e. labeled)
#     or partially-observed (i.e. unlabeled) but never mixed.
#     """

#     def __init__(self, data_x, data_y, batch_size, num_classes=2, missing_label=-1):
#         super().__init__()
#         self.data_x = data_x
#         self.data_y = data_y
#         self.batch_size = batch_size
#         self.num_classes = num_classes

#         self.unlabeled = torch.where(data_y == missing_label)[0]
#         self.num_unlabeled = self.unlabeled.size(0)
#         self.num_unlabeled_batches = math.ceil(self.num_unlabeled / self.batch_size)

#         self.labeled = torch.where(data_y != missing_label)[0]
#         self.num_labeled = self.labeled.size(0)
#         self.num_labeled_batches = math.ceil(self.num_labeled / self.batch_size)

#         assert self.data_x.size(0) == self.data_y.size(0)
#         assert len(self) > 0

#     @property
#     def size(self):
#         return self.data_x.size(0)

#     def __len__(self):
#         return self.num_unlabeled_batches + self.num_labeled_batches

#     def _sample_batch_indices(self):
#         batch_order = torch.randperm(len(self)).tolist()
#         unlabeled_idx = self.unlabeled[torch.randperm(self.num_unlabeled)]
#         labeled_idx = self.labeled[torch.randperm(self.num_labeled)]

#         slices = []

#         for i in range(self.num_unlabeled_batches):
#             _slice = unlabeled_idx[i * self.batch_size : (i + 1) * self.batch_size]
#             slices.append((_slice, False))

#         for i in range(self.num_labeled_batches):
#             _slice = labeled_idx[i * self.batch_size : (i + 1) * self.batch_size]
#             slices.append((_slice, True))

#         return slices, batch_order

#     def __iter__(self):
#         slices, batch_order = self._sample_batch_indices()

#         for i in range(len(batch_order)):
#             _slice = slices[batch_order[i]]
#             if _slice[1]:
#                 # labeled
#                 yield self.data_x[_slice[0]], nn.functional.one_hot(
#                     self.data_y[_slice[0]], num_classes=self.num_classes
#                 )
#             else:
#                 # unlabeled
#                 yield self.data_x[_slice[0]], None

def transform_numpy(data, cuda):
    if cuda==True:
        data = torch.Tensor(data).cuda()
    else:
        data = torch.Tensor(data)
    return data

def shuffer(adata):
    idx = np.random.permutation(len(adata.obs_names))
    shuffled_obs_names = adata.obs_names[idx]
    adata = adata[shuffled_obs_names]
    return adata

# class BatchSampler(Sampler):
#     def __init__(self, adata, batch_size, label_encoder = LabelEncoder()):
#         self.labels = torch.Tensor(label_encoder.fit_transform(adata.obs['batch'])).cuda()
#         self.batch_size = batch_size
#         self.num_labels = torch.unique(self.labels)

#         self.classed = [torch.where(self.labels == c)[0] for c in range(len(self.num_labels))]
#         self.counts_labels = [c.size(0) for c in self.classed]
#         self.num_class_batches = [math.ceil(indices / self.batch_size) for indices in self.counts_labels]

#     def __iter__(self):
#         iter_labels = self.labels # 可迭代的
#         indices = {} # 用于储存不同的labels对应的index
#         for c in self.num_labels.tolist():
#             indices[c] = [i for i, label in enumerate(self.labels) if label == c]
#         while len(iter_labels) > 0:
#             batch_labels = iter_labels[0].item()
#             select_indices = indices[batch_labels]
#             indices_iter = [i for i, label in enumerate(iter_labels) if label == batch_labels]

#             if len(select_indices) >= self.batch_size:
#                 yield select_indices[:self.batch_size]
#                 indices[batch_labels] = list(set(indices[batch_labels]) - set(select_indices[:self.batch_size]))
#                 iter_labels = torch.Tensor([label for i, label in enumerate(iter_labels) if i not in indices_iter[:self.batch_size]])
#             else:
#                 yield select_indices
#                 indices[batch_labels] = []
#                 iter_labels = torch.Tensor([label for i, label in enumerate(iter_labels) if i not in indices_iter])
                
#     def __len__(self):
#         return np.sum(self.num_class_batches)

class BatchSampler(Sampler):
    def __init__(self, adata, batch_size, cuda, label_encoder = LabelEncoder()):
        self.cuda = cuda
        if cuda:
            self.labels = torch.tensor(label_encoder.fit_transform(adata.obs['batch'])).cuda()
        else:
            self.labels = torch.tensor(label_encoder.fit_transform(adata.obs['batch'])).cpu()
        self.batch_size = batch_size
        self.num_labels = torch.unique(self.labels)
        self.indices = {}  # Precompute indices

        for c in self.num_labels.tolist():
            self.indices[c] = torch.where(self.labels == c)[0]

        self.num_class_batches = [math.ceil(len(indices) / self.batch_size) for indices in self.indices.values()]
        self.iter_labels = [torch.ones(n)*i for i,n in enumerate(self.num_class_batches)]
        self.iter_labels = torch.cat(self.iter_labels).int().tolist()

    def __iter__(self):
        iter_indices = self.indices.copy()
        for batch_labels in self.iter_labels:
            select_indices = iter_indices[batch_labels]
            if len(select_indices) >= self.batch_size:
                yield select_indices[:self.batch_size]
                iter_indices[batch_labels] = select_indices[self.batch_size:]
            else:
                yield select_indices
                iter_indices[batch_labels] = torch.tensor([], dtype=torch.long)  # Empty tensor

    def __len__(self):
        return np.sum(self.num_class_batches)


class scDataset(Dataset):
    def __init__(self, adata, cuda, transform=None):
        super().__init__()
        self.batch_label = pd.get_dummies(adata.obs['batch_2']).values
        self.condition_label = pd.get_dummies(adata.obs['batch']).values
        self.expr = adata.X # 注意需不需要toarray()
        self.transform = transform
        self.cuda = cuda

    def __len__(self):
        return len(self.batch_label)
    
    def __getitem__(self, index):
        expr = self.expr[index,:]
        batch_label = self.batch_label[index,:]
        condition_label = self.condition_label[index,:]
        ifcuda = self.cuda
        if self.transform:
            expr = transform_numpy(expr,ifcuda)
            batch_label = transform_numpy(batch_label,ifcuda)
            condition_label = transform_numpy(condition_label,ifcuda)
        return expr,condition_label,batch_label


# setup environment
import os
smoke_test = ('CI' in os.environ)  # for continuous integration tests

import torch
import scanpy as sc
# from scib import me
import pandas as pd
import numpy as np
import anndata as ad
import scanpy as sc
from sklearn.neighbors import NearestNeighbors
from scipy.sparse import csr_matrix

def process_batch_X(datasets):
    ########### 这个函数用于提取基因表达阵x 和 条件信息batch ########### 
    batch = datasets.obs['batch']
    batch_labels = batch.unique()
    try:
        datasets.X = datasets.X.todense() # 所有数据
        print('稀疏矩阵转为密集型矩阵！')
    except:
        pass

    if isinstance(datasets.X, csr_matrix):
        datasets.layers['X'] = datasets.X.toarray()
    else:
        datasets.layers['X'] = datasets.X
        
    x = torch.tensor(datasets.layers['X'])
    x_list = [torch.tensor(datasets.layers['X'][datasets.obs['batch'] == label]) for label in batch_labels] # 各个条件的数据

    datasets.obs['one_hot'] = -1 # 创建batch组的one-hot-vector
    for i, label in enumerate(batch_labels):
        datasets.obs.loc[datasets.obs['batch'] == label, 'one_hot'] = i
    
    one_hot = torch.tensor(datasets.obs['one_hot'].tolist()) # 将其存成tensor格式
    x = x.to(torch.float32)
    x_list = [x.to(torch.float32) for x in x_list]

    return x, x_list, batch, batch_labels, one_hot

def hyper_param(x, x_list, batch_labels):
    ########### 这个函数用于计算模型需要的一些参数，包括基因数量、条件数量以及各条件下基因表达的均值和方差 ########### 
    num_genes = x.shape[1]
    num_batch = len(batch_labels)

    x_list_mean = [x.mean(axis=0) for x in x_list]
    x_list_scale = [x.std(axis=0) for x in x_list]

    return num_genes, num_batch, x_list_mean, x_list_scale

# def Find_ari(datasets):
#     ########### 这个函数用于寻找模型的最优聚类 ########### 
#     ari_max = -1
#     index_i = 0
#     for i in range(0,30):
#         sc.tl.louvain(datasets,resolution=i*0.01,key_added='seurat_clusters')
#         ari = me.ari(datasets,'groundtruth','seurat_clusters')
#         if ari>ari_max:
#             ari_max = ari
#             index_i = i
#     return ari_max,index_i


def construct_anndata(data,columns_names,index_names,raw):
    ########### 这个函数构建新的anndata用来进行下游分析 ########### 
    DATA = pd.DataFrame(data=data, columns=columns_names,index=index_names)
    datasets = sc.AnnData(DATA)
    datasets.obs['sample_id'] = DATA.index  # 添加样本 ID 列
    datasets.var['gene_name'] = DATA.columns  # 添加基因名列

    datasets.raw = datasets
    datasets.obs = raw.obs

    datasets.layers['lognorm'] = datasets.X.copy()
    # scale数据这是用于PCA的！！
    sc.pp.scale(datasets, max_value=8)
    datasets.layers['scale'] = datasets.X.copy()
    datasets.layers['counts'] = np.expm1(datasets.layers['scale']-1).astype(int) # 反向操作
    # sc.tl.pca(datasets, svd_solver='arpack')
    # sc.pp.neighbors(datasets, n_neighbors=50, n_pcs=25)
    # sc.tl.umap(datasets)
    
    return datasets

def create_hyper(datasets,var_names,index_names,batch_size=100):
    x, x_list, batch, batch_labels, one_hot = process_batch_X(datasets)
    num_genes, num_batch, x_list_mean, x_list_scale = hyper_param(x, x_list, batch_labels)
    num_genes = x.shape[1]
    batch_size = batch_size
    hyper = {'num_genes': num_genes,
    'num_batch':num_batch,
    'batch_size':batch_size,
    'batch_labels':batch_labels,
    'one_hot':one_hot,
    'x':x,
    'x_list':x_list,
    'batch':batch,
    'var_names':var_names,
    'index_names':index_names
    }
    return hyper


def fit_knn(mat_train, mat_holdout, n_neighbors, algorithm = 'kd_tree'):
    
    # fit knn using mat_train
    # return nn indices and distances in train set for holdout set
    knn = NearestNeighbors(n_neighbors = n_neighbors, algorithm = algorithm).fit(mat_train)
    distances, indices = knn.kneighbors(mat_holdout)
    indices = indices[:,1:]
    distances = distances[:,1:]

    return indices, distances



def calc_knn_prop(knn_indices, labels_train, label_categories):

    # knn_indices: shape = (n_holdout_samples, (knn-1)), np.array
    # labels_train: shape = (n_train_samples, ), pd.object
    # label_categories: shape = (n_label_categories, ), np.array
    n = knn_indices.shape[0]
    n_category = label_categories.shape[0]
    nn_prop = np.zeros(shape = (n, n_category))

    for i in range(n):
        knn_labels = labels_train[knn_indices[i,]]
        for k in range(n_category):
            nn_prop[i, k] = sum(knn_labels == label_categories[k]) 

    nn_prop = nn_prop / knn_indices.shape[1]
    return nn_prop


def calc_oobNN(adata_orig, batch_key, condition_key, n_neighbors=15, holdout=True):
    np.random.seed(123)
    
    list_holdout = []
    for holdout_idx in np.unique(adata_orig.obs[batch_key]):
        if holdout == True:
            adata_train = adata_orig[~adata_orig.obs[batch_key].isin([holdout_idx])]
            adata_holdout = adata_orig[adata_orig.obs[batch_key].isin([holdout_idx])]
        else:
            adata_train = adata_orig
            adata_holdout = adata_orig[adata_orig.obs[batch_key].isin([holdout_idx])]
        num_cells = adata_train.obs[condition_key].value_counts().min()

    
        a_list = []
        for x in np.unique(adata_train.obs[condition_key]):
            a1 = adata_train[adata_train.obs[condition_key].isin([x])]
            random_indices = np.random.choice(a1.shape[0], size=num_cells, replace = False)
            a1 = a1[random_indices,:]
            a_list.append(a1)
 
        adata_train = ad.concat(a_list)
        adata = ad.concat([adata_train, adata_holdout])
    
        mat = sc.tl.pca(adata.X, n_comps = 20)

        mat_train = mat[:adata_train.obs.shape[0],]
        mat_holdout = mat[adata_train.obs.shape[0]:,]

    
        # fit knn
        indices, distances = fit_knn(mat_train=mat_train, mat_holdout=mat_holdout, n_neighbors=n_neighbors, algorithm = 'kd_tree')

        # compute proprotion
        labels_train = adata_train.obs[condition_key].astype('object')
        label_categories = np.unique(labels_train)
        result = calc_knn_prop(indices, labels_train, label_categories)
        knn_df = pd.DataFrame(data=result, 
                              index =  adata_holdout.obs_names,
                              columns = label_categories)
        adata_holdout.obsm['knn_prop'] = knn_df
        list_holdout.append(adata_holdout)

    res = ad.concat(list_holdout)
    return res


def calc_oobNN2(adata_orig, batch_key, n_neighbors=15):
    np.random.seed(123)
    
    list_holdout = []
    for holdout_idx in np.unique(adata_orig.obs[batch_key]):

        adata_train = adata_orig[~adata_orig.obs[batch_key].isin([holdout_idx])]
        adata_train.obs['cobatch'] = 'Rest'
        adata_holdout = adata_orig[adata_orig.obs[batch_key].isin([holdout_idx])]
        adata_holdout.obs['cobatch'] = 'Self'

        num_cells = min(adata_train.obs.shape[0],adata_holdout.obs.shape[0])

        a_list = []
        for a1 in [adata_train,adata_holdout]:
            random_indices = np.random.choice(a1.shape[0], size=num_cells, replace = False)
            a1 = a1[random_indices,:]
            a_list.append(a1)

        adata_train = ad.concat(a_list)
        adata = ad.concat([adata_train, adata_holdout])
    
        mat = sc.tl.pca(adata.X, n_comps = 20)

        mat_train = mat[:adata_train.obs.shape[0],]
        mat_holdout = mat[adata_train.obs.shape[0]:,]

    
        # fit knn
        indices, distances = fit_knn(mat_train=mat_train, mat_holdout=mat_holdout, n_neighbors=n_neighbors, algorithm = 'kd_tree')

        # compute proprotion
        labels_train = adata_train.obs['cobatch'].astype('object')
        label_categories = np.unique(labels_train)
        result = calc_knn_prop(indices, labels_train, label_categories)
        knn_df = pd.DataFrame(data = result, 
                              index =  adata_holdout.obs_names,
                              columns = label_categories)
        adata_holdout.obsm['knn_prop'] = knn_df
        list_holdout.append(adata_holdout)
    
    res = ad.concat(list_holdout)
    return res


# setup environment
import os

from . import preprocess


# various import statements
import torch
import torch.nn as nn
from torch.nn.functional import softplus, softmax
from torch.distributions import constraints

import pyro
import pyro.distributions as dist
import pyro.poutine as poutine
from pyro.distributions.util import broadcast_shape
from pyro.optim import MultiStepLR
from pyro.infer import SVI, config_enumerate, TraceEnum_ELBO, RenyiELBO
from torch.utils.data import DataLoader
import numpy as np

import time

from . import utils

####################### 构建编码解码器的函数 ###########################
# Helper for making fully-connected neural networks
def make_fc(dims):
    layers = []
    for in_dim, out_dim in zip(dims, dims[1:]):
        layers.append(nn.Linear(in_dim, out_dim))
        layers.append(nn.BatchNorm1d(out_dim))
        layers.append(nn.ReLU())
    return nn.Sequential(*layers[:-1])  # Exclude final ReLU non-linearity

# Helper for making fully-connected neural networks
def make_fc2(dims):
    layers = []
    for in_dim, out_dim in zip(dims, dims[1:]):
        layers.append(nn.Linear(in_dim, out_dim))
        layers.append(nn.ReLU())
    return nn.Sequential(*layers[:-1])  # Exclude final ReLU non-linearity

# Splits a tensor in half along the final dimension
def split_in_half(t):
    return t.reshape(t.shape[:-1] + (2, -1)).unbind(-2)

# Helper for broadcasting inputs to neural net
def broadcast_inputs(input_args):
    shape = broadcast_shape(*[s.shape[:-1] for s in input_args]) + (-1,)
    input_args = [s.expand(shape) for s in input_args]
    return input_args

####################### 完成需要的编码-解码器 ###########################
# decoder:
# total_special_decoder(z_total,z_special)
# x_decoder(z)
# encode:
# z_encoder(x)
# total_special_encoder(z)

######################## Decoder Part #################################
# Used in parameterizing p(z |z_total,z_special)
class ZSDecoder(nn.Module):
    def __init__(self, num_batch, z_special_dim, z_total_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [num_batch + z_total_dim] + hidden_dims1 + [2 * z_special_dim]
        self.fc = make_fc(dims)

    def forward(self, t ,z_total):
        z_t = broadcast_inputs([z_total, t])
        z_t = torch.cat(z_t, dim=-1)
        # We reshape the input to be two-dimensional so that nn.BatchNorm1d behaves correctly
        _z_t = z_t.reshape(-1, z_t.size(-1))
        hidden = self.fc(_z_t)
        # If the input was three-dimensional we now restore the original shape
        hidden = hidden.reshape(z_t.shape[:-1] + hidden.shape[-1:])
        loc, scale = split_in_half(hidden)
        scale = softplus(scale)
        return loc, scale


# Used in parameterizing p(x |z)
class XDecoder(nn.Module):
    # This __init__ statement is executed once upon construction of the neural network.
    # Here we specify that the neural network has input dimension z2_dim
    # and output dimension 2 * num_genes.
    def __init__(self, num_genes, z_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_dim] + hidden_dims1 + [2 * num_genes]
        self.fc = make_fc(dims)


    def forward(self, z):
        loc, scale = split_in_half(self.fc(z))
        # Note that mu is normalized so that total count information is
        # encoded by the latent variable ℓ.
        scale = softplus(scale)
        return loc, scale
    
class XDecoder_possion(nn.Module):
    # This __init__ statement is executed once upon construction of the neural network.
    # Here we specify that the neural network has input dimension z2_dim
    # and output dimension num_genes.
    def __init__(self, num_genes, z_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_dim] + hidden_dims1 + [num_genes]
        self.fc = make_fc(dims)


    def forward(self, z):
        u = self.fc(z)
        # Note that mu is normalized so that total count information is
        # encoded by the latent variable ℓ.
        u = softplus(u)
        return u


# Used in parameterizing q(z_total | z) and q(z_special | z)
class ZDecoder(nn.Module):
    def __init__(self, z_dim, z_total_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_total_dim] + hidden_dims1  + [2 * z_dim]
        self.fc = make_fc2(dims)

    def forward(self, z_total):
        loc, scale = split_in_half(self.fc(z_total))
        # Note that mu is normalized so that total count information is
        # encoded by the latent variable ℓ.
        scale = softplus(scale)
        return loc, scale

# Used in parameterizing q(z_total | z) and q(z_special | z)
class Z2Decoder(nn.Module):
    def __init__(self, z_dim, z_total_dim, z_special_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_total_dim + z_special_dim] + hidden_dims1  + [2 * z_dim]
        self.fc = make_fc2(dims)

    def forward(self, z_total, z_special):
        z_total_special = torch.cat([z_total,z_special], dim=-1)
        # We reshape the input to be two-dimensional so that nn.BatchNorm1d behaves correctly
        _z_total_special = z_total_special.reshape(-1, z_total_special.size(-1))
        hidden = self.fc(_z_total_special)
        # If the input was three-dimensional we now restore the original shape
        hidden = hidden.reshape(_z_total_special.shape[:-1] + hidden.shape[-1:])
        loc, scale = split_in_half(hidden)
        # Here and elsewhere softplus ensures that scale is positive. Note that we generally
        # expect softplus to be more numerically stable than exp.
        scale = softplus(scale)
        return loc, scale


######################## Encoder Part #################################

class TSEncoder(nn.Module):
    def __init__(self, z_dim, z_special_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_dim] + hidden_dims1 + [2 * z_special_dim]
        self.fc = make_fc(dims)
        self.z_special_dim = z_special_dim

    def forward(self, z):
        # Transform the counts x to log space for increased numerical stability.
        # Note that we only use this transformation here; in particular the observation
        # distribution in the model is a proper count distribution.
        loc, scale = split_in_half(self.fc(z))
        # Here and elsewhere softplus ensures that scale is positive. Note that we generally
        # expect softplus to be more numerically stable than exp.
        scale = softplus(scale)
        return loc, scale



class ZSEncoder(nn.Module):
    def __init__(self, num_batch, z_special_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_special_dim] + hidden_dims1 + [num_batch]
        self.fc = make_fc(dims)

    def forward(self, z_special):
        # Transform the counts x to log space for increased numerical stability.
        # Note that we only use this transformation here; in particular the observation
        # distribution in the model is a proper count distribution.
        logits = self.fc(z_special)
        return logits

class ZS2Encoder(nn.Module):
    def __init__(self, num_batch, z_special_dim, z_total_dim, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [z_special_dim + num_batch] + hidden_dims1 + [2 * z_total_dim]
        self.fc = make_fc(dims)
        self.z_special_dim = z_special_dim

    def forward(self, z_special, t):
        # This broadcasting is necessary since Pyro expands y during enumeration (but not z2)
        zs_t = broadcast_inputs([z_special, t])
        zs_t = torch.cat(zs_t, dim=-1)
        # We reshape the input to be two-dimensional so that nn.BatchNorm1d behaves correctly
        _zs_t = zs_t.reshape(-1, zs_t.size(-1))
        hidden = self.fc(_zs_t)
        # If the input was three-dimensional we now restore the original shape
        hidden = hidden.reshape(zs_t.shape[:-1] + hidden.shape[-1:])
        loc, scale = split_in_half(hidden)
        scale = softplus(scale)
        return loc, scale

# Used in parameterizing q(z_total,t | x)
class ZEncoder(nn.Module):
    def __init__(self, z_dim, num_genes, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [num_genes] + hidden_dims1 + [2 * z_dim]
        self.fc = make_fc(dims)

    def forward(self, x):
        # Transform the counts x to log space for increased numerical stability.
        # Note that we only use this transformation here; in particular the observation
        # distribution in the model is a proper count distribution.
        h1, h2 = split_in_half(self.fc(x))
        loc, scale = h1, softplus(h2)
        return loc, scale
    
class ZEncoder_possion(nn.Module):
    def __init__(self, z_dim, num_genes, hidden_dims1, hidden_dims2):
        super().__init__()
        dims = [num_genes] + hidden_dims1 + [z_dim]
        self.fc = make_fc(dims)

    def forward(self, x):
        # Transform the counts x to log space for increased numerical stability.
        # Note that we only use this transformation here; in particular the observation
        # distribution in the model is a proper count distribution.
        u = self.fc(x)
        u= softplus(u)
        return u

####################### 构建pyro模型 ###########################
class iPerturb_model_init(nn.Module):
    def __init__(self, num_genes, num_batch, latent_dim1=20, latent_dim2=10, latent_dim3=10, alpha=1e-5, scale_factor=1.0, mode = 'lognorm'):
        self.num_genes = num_genes
        self.num_batch = num_batch

        # This is the latent_dim dimension of both z, z_total, z_special
        self.latent_dim1 = latent_dim1
        self.latent_dim2 = latent_dim2
        self.latent_dim3 = latent_dim3

        # This hyperparameter controls the strength of the auxiliary classification loss
        self.alpha = alpha
        self.scale_factor = scale_factor
        self.mode = mode

        super().__init__()

        # Setup the various neural networks used in the model and guide
        self.z_decoder = ZDecoder(z_dim=self.latent_dim1, z_total_dim=self.latent_dim2,
                                    hidden_dims1=[100], hidden_dims2=[50])

        self.z2_decoder = Z2Decoder(z_dim=self.latent_dim1, z_total_dim=self.latent_dim2,
                                    z_special_dim=self.latent_dim3, hidden_dims1=[100], hidden_dims2=[50])

        self.zs_decoder = ZSDecoder(num_batch=self.num_batch ,z_special_dim=self.latent_dim3, z_total_dim=self.latent_dim2, hidden_dims1=[100], hidden_dims2=[50])

        if mode == 'lognorm':
            self.x_decoder = XDecoder(num_genes=num_genes, hidden_dims1=[100], hidden_dims2=[50], z_dim=self.latent_dim1)
        elif mode == 'possion':
            self.x_decoder = XDecoder_possion(num_genes=num_genes, hidden_dims1=[100], hidden_dims2=[50], z_dim=self.latent_dim1)

        self.ts_encoder = TSEncoder(z_dim=self.latent_dim1,
                                    z_special_dim=self.latent_dim3 ,hidden_dims1=[100], hidden_dims2=[50])

        self.zs_encoder = ZSEncoder(num_batch=self.num_batch, z_special_dim=self.latent_dim3 ,hidden_dims1=[100], hidden_dims2=[50])


        self.z_encoder = ZEncoder(z_dim=self.latent_dim1, num_genes=num_genes, hidden_dims1=[100], hidden_dims2=[50])


        self.zs2_encoder = ZS2Encoder(  num_batch=self.num_batch, 
                                        z_total_dim=self.latent_dim2, z_special_dim=self.latent_dim3 ,hidden_dims1=[100], hidden_dims2=[50])

        self.epsilon = 0.006

    def model(self, x, one_hot):
        # Register various nn.Modules (i.e. the decoder/encoder networks) with Pyro
        pyro.module("iPerutrb_model", self)

        # This gene-level parameter modulates the variance of the observation distribution
        theta = pyro.param("inverse_dispersion", 10.0 * x.new_ones(self.num_genes),
                        constraint=constraints.positive)

        mode = self.mode
        # We scale all sample statements by scale_factor so that the ELBO loss function
        # is normalized wrt the number of datapoints and genes.
        # This helps with numerical stability during optimization.
        with pyro.plate("batch", len(x)), poutine.scale(scale=self.scale_factor):
            z_total = pyro.sample("z_total", dist.Normal(0, x.new_ones(self.latent_dim2)).to_event(1))
            t_logit = pyro.sample("t_logit", dist.Dirichlet(x.new_ones(self.num_batch)))
            t = pyro.sample("t", dist.OneHotCategorical(logits=t_logit), obs=one_hot)

            z_special_loc, z_special_scale = self.zs_decoder(z_total, t_logit) # t_logit or t_smooth
            z_special = pyro.sample("z_special", dist.Normal(z_special_loc, z_special_scale).to_event(1))

            z_loc, z_scale = self.z2_decoder(z_total, z_special)
            z = pyro.sample("z", dist.Normal(z_loc, z_scale).to_event(1))

            if mode == 'lognorm':
                x_loc, x_scale = self.x_decoder(z)
                x_scale = softplus(x_scale + theta + self.epsilon)
                x = pyro.sample("x", dist.Normal(x_loc, x_scale).to_event(1), obs=x)
            elif mode == 'possion':
                xu = self.x_decoder(z)
                x = pyro.sample("x", dist.Poisson(xu).to_event(1), obs=x)

    # The guide specifies the variational distribution
    def guide(self, x, one_hot):
        pyro.module("iPerutrb_model", self)

        with pyro.plate("batch", len(x)), poutine.scale(scale=self.scale_factor):
            z_loc, z_scale = self.z_encoder(x)
            z = pyro.sample("z", dist.Normal(z_loc, z_scale).to_event(1))

            z_special_loc, z_special_scale = self.ts_encoder(z)
            z_special = pyro.sample("z_special", dist.Normal(z_special_loc, z_special_scale).to_event(1))

            t_logit = self.zs_encoder(z_special)

            t_dist = dist.OneHotCategorical(logits=t_logit)
            if one_hot is None:
                t = pyro.sample("t", t_dist)
            else:
                # x is labeled so add a classification loss term
                # (this way q(y|z2) learns from both labeled and unlabeled data)
                classification_loss = t_dist.log_prob(one_hot)
                # Note that the negative sign appears because we're adding this term in the guide
                # and the guide log_prob appears in the ELBO as -log q
                pyro.factor("classification_loss", -self.alpha * classification_loss, has_rsample=False)

            z_total_loc, z_total_scale = self.zs2_encoder(z_special, t_logit)
            z_total = pyro.sample("z_total", dist.Normal(z_total_loc, z_total_scale).to_event(1))

    
# 用来记录pyro模型迭代的重要参数，并进行模型初始化
def model_init_(hyper, latent_dim1, latent_dim2, latent_dim3, optimizer, lr, gamma, milestones, set_seed=123, cuda = False, alpha=0, mode = 'lognorm'):
    num_genes = hyper['num_genes']
    num_batch = hyper['num_batch']
    scale_factor = 1.0 / (hyper['batch_size'] * hyper['num_genes'])
    # Clear Pyro param store so we don't conflict with previous
    # # training runs in this session
    # pyro.clear_param_store()
    # Fix random number seed
    pyro.util.set_rng_seed(set_seed)
    # Enable optional validation warnings
    pyro.enable_validation(True)
    # Instantiate instance of model/guide and various neural networks
    iPerturb_model = iPerturb_model_init(num_genes=num_genes,num_batch=num_batch,latent_dim1=latent_dim1, latent_dim2=latent_dim2, latent_dim3=latent_dim3, scale_factor=scale_factor,alpha=alpha, mode = mode)
    if cuda == True:
        iPerturb_model = iPerturb_model.cuda()
    # Setup an optimizer (Adam) and learning rate scheduler.
    # We start with a moderately high learning rate (0.006) and
    # reduce by a factor of 5 after 20 epochs.
    scheduler = MultiStepLR({'optimizer': optimizer,
                            'optim_args': {'lr': lr},
                            'gamma': gamma, 'milestones': milestones})

    # Tell Pyro to enumerate out y when y is unobserved.
    # (By default y would be sampled from the guide)
    guide = config_enumerate(iPerturb_model.guide, "parallel", expand=True)
    # Setup a variational objective for gradient-based learning.
    # Note we use TraceEnum_ELBO in order to leverage Pyro's machinery
    # for automatic enumeration of the discrete latent variable y.


    elbo = TraceEnum_ELBO(strict_enumeration_warning=False)
    # elbo = RenyiELBO(strict_enumeration_warning=False)
    
    svi = SVI(iPerturb_model.model, guide, scheduler, elbo)
    return svi, scheduler, iPerturb_model

## 利用pyro进行模型迭代
def RUN_model(epochs,svi,scheduler,dataloader):
    pyro.clear_param_store()
    start_time = time.time()
    for epoch in range(epochs):
        losses = []

        # Take a gradient step for each mini-batch in the dataset
        for x,one_hot,_ in dataloader:
            if one_hot is not None:
                one_hot = one_hot.type_as(x)

            loss = svi.step(x,one_hot)
            losses.append(loss)

        # Tell the scheduler we've done one epoch.
        scheduler.step()

        print("[Epoch %02d]  Loss: %.5f" % (epoch, np.mean(losses)))
    end_time = time.time()
    execution_time = end_time - start_time
    print(f"Finished training! Time:{execution_time}.")

# 总的RUN
def RUN(datasets,iPerturb_model,svi,scheduler,epochs,hyper,raw,cuda,batch_size=100,if_likelihood=False):
    sc_datasets = preprocess.scDataset(datasets,cuda=cuda,transform=preprocess.transform_numpy)
    batch_sampler = preprocess.BatchSampler(datasets,batch_size=batch_size,cuda=cuda)

    dataloader = DataLoader(sc_datasets, batch_sampler=batch_sampler)
    dataloader2 = DataLoader(sc_datasets, batch_size=batch_size, shuffle=True)

    # training runs in this session
    RUN_model(epochs,svi,scheduler,dataloader)
    # Put the neural networks in evaluation mode (needed because of batch norm)
    iPerturb_model.eval()
    if cuda:
        # z
        z_pred = iPerturb_model.z_encoder(hyper['x'].cuda())[0]
    else:
        # z
        z_pred = iPerturb_model.z_encoder(hyper['x'])[0]
    # 重构
    mode = iPerturb_model.mode
    if mode == 'lognorm':
        x_pred = iPerturb_model.x_decoder(z_pred)[0]
    elif mode == 'possion':
        x_pred = iPerturb_model.x_decoder(z_pred)
    z_s = iPerturb_model.ts_encoder(z_pred)[0]
    t_logits = iPerturb_model.zs_encoder(z_s)
    z_t = iPerturb_model.zs2_encoder(z_s, t_logits)[0]

    x_pred = x_pred.data.cpu().numpy()
    z_pred = z_pred.data.cpu().numpy()
    zs_pred = z_s.data.cpu().numpy()
    zt_pred = z_t.data.cpu().numpy()
    t_logits = softmax(t_logits,dim=-1).data.cpu().numpy()
    likelihood = np.argmax(t_logits, axis=-1)

    # # likelihood
    # if if_likelihood:
    #     RUN_model(epochs,svi,scheduler,dataloader2)
    #     iPerturb_model.eval()
    #     if cuda:
    #         # z
    #         z_pred_ = iPerturb_model.z_encoder(hyper['x'].cuda())[0]
    #     else:
    #         # z
    #         z_pred_ = iPerturb_model.z_encoder(hyper['x'])[0]
    #     zs_pred_ = iPerturb_model.ts_encoder(z_pred_)[0]
    #     t_logits = iPerturb_model.zs_encoder(zs_pred_)
    #     t_logits = softmax(t_logits,dim=-1).data.cpu().numpy()
    #     likelihood = np.argmax(t_logits, axis=-1)


    # 创建 AnnData 对象
    datasets = utils.construct_anndata(data = x_pred, columns_names = hyper['var_names'], index_names = hyper['index_names'], raw=raw)  
        
    datasets.obsm['x_emb'] = z_pred
    datasets.obsm['zt_emb'] = zt_pred
    datasets.obsm['zs_emb'] = zs_pred

    datasets.obsm['t_logits'] = t_logits
    datasets.obs['likelihood'] = likelihood
    datasets.obs['t_logits'] = datasets.obsm['t_logits'][:,0]
    
    if if_likelihood:
        datasets.obsm['t_logits'] = t_logits
        datasets.obs['likelihood'] = likelihood
        datasets.obs['t_logits'] = datasets.obsm['t_logits'][:,0]

    # ari_max,index_i = utils.Find_ari(datasets)
    # print(f'第一层重构已完成,它的ari是{ari_max},它的resolution是{index_i}.')


    return x_pred,datasets


# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import io

with io.open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with io.open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read().splitlines()

setup(
    name="iperturb",  # 包的名字
    version="0.2.7",  # 初始版本
    author="Billy Chen",
    author_email="cyz2022@stu.xjtu.edu.cn",
    description="Atlas-level data integration in multi-condition single-cell genomics",  # 简短描述
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BillyChen123/iPerturb",  # 项目的主页
    packages=find_packages(),  # 自动找到包
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    tests_require=[
        'pytest',
    ],
)
